extends Node

var has_available: bool = true

func reserve_printer() -> void:
	if not has_available:
		return
	Events.printer_reserved.emit()

func check_target() -> void:
	if target == null:
		return

func check_queue() -> void:
	if queue.is_empty():
		return

func check_node() -> void:
	if not is_instance_valid(node):
		return
