"""GodotIQ Godot editor addon — package data anchor.

This package's directory contains the Godot addon files (plugin.cfg, *.gd)
bundled via hatchling force-include. Use Path(__file__).parent to locate them.
"""
