extends Node2D

const ANIM_MAP := {
	"idle": "idle",
	"walk": "walk",
	"attack": "attack",
	"special": "special_move",
}

const LOOPING_ANIMS := ["idle", "walk"]

var anim_player: AnimationPlayer

func _ready() -> void:
	anim_player = $AnimationPlayer
	anim_player.play("idle")

func attack() -> void:
	anim_player.play("attack")

func special() -> void:
	anim_player.play("special_move")
