"""Tests for animation tool server registration."""

from __future__ import annotations

import pytest

from godotiq.server import mcp


def test_import_animation_module():
    """Importing godotiq.tools.animation should not raise."""
    import godotiq.tools.animation  # noqa: F401


class TestAnimationToolRegistration:
    """Verify animation tools are registered on the MCP server."""

    @pytest.mark.asyncio
    async def test_mcp_has_animation_info_tool(self) -> None:
        """The mcp server instance should have godotiq_animation_info registered."""
        tools = await mcp.list_tools()
        tool_names = [t.name for t in tools]
        assert "godotiq_animation_info" in tool_names

    @pytest.mark.asyncio
    async def test_mcp_has_animation_audit_tool(self) -> None:
        """The mcp server instance should have godotiq_animation_audit registered."""
        tools = await mcp.list_tools()
        tool_names = [t.name for t in tools]
        assert "godotiq_animation_audit" in tool_names
