extends Node3D

const ANIM_MAP := {
	"idle": "NlaTrack_001",
	"walk": "NlaTrack_002",
	"run": "NlaTrack_003",
}

const LOOPING_ANIMS := ["idle", "walk"]

var anim_player: AnimationPlayer

func _ready() -> void:
	anim_player = $AnimationPlayer
	anim_player.play("NlaTrack_001")

func play_animation(anim_name: String) -> void:
	var track_name = ANIM_MAP.get(anim_name, "")
	anim_player.play(track_name)
