class_name ChainedManagerClass
extends Node

signal task_completed(task_id: String)

func accept_item(item_id: String) -> void:
	_process_item(item_id)

func _process_item(item_id: String) -> void:
	_create_task(item_id)

func _create_task(item_id: String) -> void:
	task_completed.emit(item_id)
